unit Imager;

interface

uses
  {$IFDEF WIN32}
  Windows,
  {$ELSE}
  WinTypes, WinProcs, SysUtils,
  {$ENDIF}
  Classes, Messages, Graphics, Controls, Menus;

{ TSmartImage }

type
  TSmartImage = class(TGraphicControl)
  private
    FPicture: TPicture;
    FBitmap: TBitmap;
    FAutoSize: Boolean;
    FStretch: Boolean;
    FMaintainAspect: Boolean;
    FTransparent: Boolean;
    FCenter: Boolean;
    FReserved: Byte;
    function GetCanvas: TCanvas;
    procedure PictureChanged(Sender: TObject);
    procedure SetAutoSize(Value: Boolean);
    procedure SetBitmap(Value: TBitmap);
    procedure SetCenter(Value: Boolean);
    procedure SetPicture(Value: TPicture);
    procedure SetStretch(Value: Boolean);
    procedure SetMaintainAspect(Value: Boolean);
    procedure SetTransparent(Value: Boolean);
    procedure CMColorChanged(var Message: TMessage);
      message CM_COLORCHANGED;
    procedure CMSysColorChange(var Message: TMessage);
      message CM_SYSCOLORCHANGE;
    procedure CMParentColorChanged(var Message: TMessage);
      message CM_PARENTCOLORCHANGED;
    procedure ResetBitmap;
  protected
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    procedure ClearBitmap;
    function GetPalette: HPALETTE; override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadFromFile(const Filename: string);
    property Canvas: TCanvas read GetCanvas;
  published
    property Align;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default False;
    property Center: Boolean read FCenter write SetCenter default False;
    { property Color; }
    property DragCursor;
    property DragMode;
    property Enabled;
    property ParentShowHint;
    { property ParentColor; }
    property Picture: TPicture read FPicture write SetPicture;
    property PopupMenu;
    property ShowHint;
    property Stretch: Boolean read FStretch write SetStretch default False;
    property MaintainAspect: Boolean read FMaintainAspect
      write SetMaintainAspect default True;
    property Transparent: Boolean read FTransparent write SetTransparent
      default False;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

{ Procedures }

function RectFitInto(Fit, Into: TRect; Center: Boolean): TRect;
function IconToBmp(const AIcon: TIcon; AColor: TColor): TBitmap;

{ Variables }

const ShowProgressBar: Boolean = True;

implementation

uses
  {$IFDEF WIN32} NViewLib, {$ENDIF}
  Consts;

{ Procedures }

function RectFitInto(Fit, Into: TRect; Center: Boolean): TRect;
var
  X, Y: Integer;
  R: Real;
  I: Integer;
begin
  Result := Into;

  OffsetRect(Fit, -Fit.Left, -Fit.Top);
  if (Fit.Right = 0) or (Fit.Bottom = 0) then Exit;

  X := Into.Left;  Y := Into.Top;
  OffsetRect(Into, -X, -Y);
  if (Into.Right = 0) or (Into.Bottom = 0) then Exit;

  R := Fit.Right / Fit.Bottom;
  I := Round(Into.Right / R);
  if (I > Into.Bottom) then
  begin
    Fit.Bottom := Into.Bottom;
    Fit.Right := Round(Into.Bottom * R);
    if Center then
      OffsetRect(Fit, (Into.Right - Fit.Right) div 2, 0);
  end
  else
  begin
    Fit.Right := Into.Right;
    Fit.Bottom := I;
    if Center then
      OffsetRect(Fit, 0, (Into.Bottom - Fit.Bottom) div 2);
  end;

  OffsetRect(Fit, X, Y);  Result := Fit;
end;

{$IFDEF WIN32}
function IconToBmp(const AIcon: TIcon; AColor: TColor): TBitmap;
var
  Bitmap: TBitmap;
  ANDBmp: TBitmap;
  i, j: Integer;
  IconInfo: TIconInfo;
  OldMaskColor: TColor;
  NewMaskColor: TColor;
begin
  Result := nil;
  if (AIcon = nil) or (AIcon.Empty) then Exit;

  Bitmap := TBitmap.Create;
  Bitmap.Width := AIcon.Width;  Bitmap.Height := AIcon.Height;
  ANDBmp := TBitmap.Create;
  try
    GetIconInfo(AIcon.Handle, IconInfo);
    ANDBmp.Handle := IconInfo.hbmMask;
    Bitmap.Width := ANDBmp.Width;
    Bitmap.Height := ANDBmp.Height + 1;
    Bitmap.Canvas.Draw(0, 0, AIcon);

    Bitmap.Canvas.Pen.Color := AColor;
    for j := 0 to ANDBmp.Height - 1 do
    begin
      OldMaskColor := clRed;
      Bitmap.Canvas.MoveTo(0, j);
      NewMaskColor := clWhite;

      for i := 0 to ANDBmp.Width - 1 do
      begin
        NewMaskColor := GetPixel(ANDBmp.Canvas.Handle, i, j);
        if (NewMaskColor <> OldMaskColor) then
        begin
          OldMaskColor := NewMaskColor;
          if (NewMaskColor = clWhite) then
            Bitmap.Canvas.MoveTo(i, j)
          else
            Bitmap.Canvas.LineTo(i, j);
        end;
      end;

      if (NewMaskColor = clWhite) then
        Bitmap.Canvas.LineTo(ANDBmp.Width, j);
    end;

    Bitmap.Canvas.MoveTo(0, Bitmap.Height - 1);
    Bitmap.Canvas.LineTo(Bitmap.Width, Bitmap.Height - 1);
  finally
    ANDBmp.Free;
  end;

  Result := Bitmap;
end;
{$ELSE}
function IconToBmp(const AIcon: TIcon; AColor: TColor): TBitmap;
var  BMP: TBitmap;
begin
  Result := nil;
  if (AIcon = nil) or (AIcon.Empty) then Exit;

  BMP := TBitmap.Create;
  BMP.Width := AIcon.Width;  BMP.Height := AIcon.Height;
  with BMP.Canvas do
  begin
    Brush.Color := AColor;
    FillRect(Rect(0, 0, AIcon.Width, AIcon.Height));
    Draw(0, 0, AIcon);
  end;
  Result := BMP;
end;
{$ENDIF}

{ TSmartImage }

constructor TSmartImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBitmap := TBitmap.Create;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  FMaintainAspect := True;
  Height := 105;
  Width := 105;
  inherited ParentColor := True;
end;

destructor TSmartImage.Destroy;
begin
  FPicture.Free;
  FBitmap.Free;
  inherited Destroy;
end;

procedure TSmartImage.LoadFromFile(const FileName: string);
begin
  {$IFDEF WIN32}
  case NVLGetType(FileName) of
    nvlJPG: Picture.Bitmap := NVLLoadJPG(FileName, ShowProgressBar);
    nvlGIF: Picture.Bitmap := NVLLoadGIF(FileName, ShowProgressBar);
  else
    Picture.LoadFromFile(FileName);
  end;
  {$ELSE}
  Picture.LoadFromFile(FileName);
  {$ENDIF}
end;

function TSmartImage.GetPalette: HPALETTE;
begin
  Result := 0;
  if (FPicture.Graphic is TBitmap) then
    if Bitmap.Empty then
      Result := TBitmap(FPicture.Graphic).Palette
    else
      Result := Bitmap.Palette;
end;

procedure TSmartImage.Paint;
var
  Dest: TRect;
  W, H: Integer;
begin
  if csDesigning in ComponentState then
    with inherited Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;

  if Bitmap.Empty then
    begin  W := Picture.Width;  H := Picture.Height;  end
  else
    begin  W := Bitmap.Width;  H := Bitmap.Height;  end;

  if Stretch then
  begin
    if MaintainAspect then
      Dest := RectFitInto(Rect(0, 0, W, H), ClientRect, Center)
    else
      Dest := ClientRect;
  end
  else
    if Center then
      Dest := Bounds((Width - W) div 2, (Height - H) div 2, W, H)
    else
      Dest := Rect(0, 0, W, H);

  with inherited Canvas do
    if Bitmap.Empty then
      StretchDraw(Dest, Picture.Graphic)
    else
      StretchDraw(Dest, Bitmap);
end;

procedure TSmartImage.ClearBitmap;
var  BMP: TBitmap;
begin
  BMP := TBitmap.Create;
  FBitmap.Assign(BMP);
  BMP.Free;
end;

procedure TSmartImage.ResetBitmap;
begin
  if ((Picture.Graphic is TBitmap) and Transparent) then
    Bitmap := Picture.Bitmap
  else
    if (Picture.Graphic is TIcon) then
      FBitmap.Assign(IconToBmp(Picture.Icon, Color));
end;

procedure TSmartImage.CMColorChanged(var Message: TMessage);
begin  inherited;  ResetBitmap;  end;

procedure TSmartImage.CMSysColorChange(var Message: TMessage);
begin  inherited;  ResetBitmap;  end;

procedure TSmartImage.CMParentColorChanged(var Message: TMessage);
begin  inherited;  ResetBitmap;  end;

function TSmartImage.GetCanvas: TCanvas;
var  BMP: TBitmap;
begin
  if not Bitmap.Empty then
    Result := Bitmap.Canvas
  else
  begin
    if (Picture.Graphic = nil) then
    begin
      BMP := TBitmap.Create;
      try
        BMP.Width := Width;
        BMP.Height := Height;
        Picture.Graphic := BMP;
      finally
        BMP.Free;
      end;
    end;
    if (Picture.Graphic is TBitmap) then
      Result := TBitmap(Picture.Graphic).Canvas
    else
      raise EInvalidOperation.Create(LoadStr(SImageCanvasNeedsBitmap));
  end;
end;

procedure TSmartImage.SetAutoSize(Value: Boolean);
begin
  FAutoSize := Value;
  PictureChanged(Self);
end;

procedure TSmartImage.SetCenter(Value: Boolean);
begin
  if (FCenter <> Value) then
  begin
    FCenter := Value;
    Invalidate;
  end;
end;

procedure TSmartImage.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TSmartImage.SetBitmap(Value: TBitmap);
var
  BMP: TBitmap;
  Dest: TRect;
begin
  BMP := TBitmap.Create;
  if Transparent and (not Value.Empty) then
  begin
    Dest := Rect(0, 0, Value.Width, Value.Height);
    BMP.Width := Dest.Right;  BMP.Height := Dest.Bottom;
    with BMP.Canvas do
    begin
      Brush.Color := Color;
      BrushCopy(Dest, Value, Dest, Value.Canvas.Pixels[0, 0]);
    end;
  end;
  Bitmap.Assign(BMP);
  BMP.Free;
end;

procedure TSmartImage.SetStretch(Value: Boolean);
begin
  if (Value <> FStretch) then
  begin
    FStretch := Value;
    Invalidate;
  end;
end;

procedure TSmartImage.SetMaintainAspect(Value: Boolean);
begin
  if (Value <> FMaintainAspect) then
  begin
    FMaintainAspect := Value;
    Invalidate;
  end;
end;

procedure TSmartImage.SetTransparent(Value: Boolean);
begin
  if (Picture.Graphic is TIcon) then Exit;
  FTransparent := Value;
  if (Picture.Graphic is TBitmap) then
    if Value then
      Bitmap := Picture.Bitmap
    else
      ClearBitmap;
  Invalidate;
end;

procedure TSmartImage.PictureChanged(Sender: TObject);
begin
  if AutoSize and (Picture.Width > 0) and (Picture.Height > 0) then
    SetBounds(Left, Top, Picture.Width, Picture.Height);
  if (Picture.Graphic is TBitmap) and (Picture.Width = Width) and
    (Picture.Height = Height) then
    ControlStyle := ControlStyle + [csOpaque] else
    ControlStyle := ControlStyle - [csOpaque];

  if (Picture.Graphic is TBitmap) and Transparent then
    Bitmap := Picture.Bitmap
  else
    if (Picture.Graphic is TIcon) then
    begin
      FBitmap.Assign(IconToBmp(Picture.Icon, Color));
      FTransparent := True;
    end
    else
      ClearBitmap;

  Invalidate;
end;

end.
